function modReset()

    -- Tool
    regSetBool('tool.highlightControllers', true)
    regSetBool('tool.showTrainingInfo'    , true)
    regSetBool('tool.showDebugMessage'    , true)
    regSetBool('tool.debugMode'           , false)
    regSetBool('tool.showAIInfo'          , true)
    regSetBool('tool.createActiveAI'      , false)
    regSetBool('tool.randomAddDuringTrain', true)

    -- Neural Network
    regSetString('neural_network.objective'   , "balance")
    local defaultNN = ai.neural_network.GetDefaultNN("balance")
    local defaultNNStr = ai.neural_network.ToJson(defaultNN)
    regSetString('neural_network.default'     , defaultNNStr)
    regSetString('neural_network.current'     , defaultNNStr)
    regSetString('neural_network.loadSelected', "none")
    local savedNeuralNetworks = regListKeys('neural_network.saved')
    if #savedNeuralNetworks > 0 then
        regSetString('neural_network.loadSelected', savedNeuralNetworks[1])
    end
    regSetString('neural_network.saveName'     , "")
    regSetInt('neural_network.accelRate'       , 1)
    regSetInt('neural_network.distToPlayer'    , 3)
    regSetInt('neural_network.scanRange'       , 10)
    regSetBool('neural_network.inactiveOnError', true)
    regSetBool('neural_network.inactiveOnBreak', true)

    -- Training
    regSetInt('training.batch_size'     , 32)
    regSetInt('training.epsilonDecrease', 8)
    regSetBool('training.batchMode'     , false)
 
    -- Genetic Algorithm
    regSetInt('genetic_algo.generationDuration', 5)
    regSetInt('genetic_algo.keepTopPercent'    , 50)
    regSetString('genetic_algo.selectionType'  , "linear")
    regSetInt('genetic_algo.crossoverPercent'  , 100)
    regSetString('genetic_algo.crossoverType'  , "slice-1")
    regSetInt('genetic_algo.mutationRate'      , 5)
    regSetInt('genetic_algo.mutationMin'       , -2)
    regSetInt('genetic_algo.mutationMax'       , 2)
    regSetBool('genetic_algo.trainDuringGen'   , true)
    regSetBool('genetic_algo.mutationMultiply' , true)
    regSetBool('genetic_algo.unbreakableTrain' , false)

    -- Key bindings
    regInitKey('key.attach'  , "lmb")
    regInitKey('key.copy'    , "mmb")
    regInitKey('key.activate', "x")
    regInitKey('key.convert' , "c")
    regInitKey('key.start'   , "r")
    regInitKey('key.pause'   , "e")
    regInitKey('key.remove'  , "k")
    regInitKey('key.options' , "o")
    
end

function getMajorVersion(versionString)
    versionString = versionString or "v0"
    local dotIndex = string.find(versionString, '.', 1, true) or #versionString
    return string.sub(versionString, 2, dotIndex)
end

function checkRegInitialized()
    local savedVersion = regGetString('savegame.mod.verison')
    local currentVersion = "v1.0.1"
    -- Only reset mod if major version changed
    if getMajorVersion(savedVersion) ~= getMajorVersion(currentVersion) then
        modReset()
    end
    regSetString('savegame.mod.verison', currentVersion)
end

function regGetFloat(path)
    local p = 'savegame.mod.' .. path
    return GetFloat(p)
end
function regSetFloat(path, value)
    local p = 'savegame.mod.' .. path
    SetFloat(p, value)
end

function regGetInt(path)
    local p = 'savegame.mod.' .. path
    return GetInt(p)
end
function regSetInt(path, value)
    local p = 'savegame.mod.' .. path
    SetInt(p, value)
end

function regGetBool(path)
    local p = 'savegame.mod.' .. path
    return GetBool(p)
end
function regSetBool(path, value)
    local p = 'savegame.mod.' .. path
    SetBool(p, value)
end

function regGetString(path)
    local p = 'savegame.mod.' .. path
    return GetString(p)
end
function regSetString(path, value)
    local p = 'savegame.mod.' .. path
    SetString(p, value)
end

function regInitKey(path, key)
    if regGetString(path) == "" then
        regSetString(path, key)
    end
end

function regListKeys(path)
    local p = 'savegame.mod.' .. path
    return ListKeys(p)
end

function regClearKey(path)
    local p = 'savegame.mod.' .. path
    return ClearKey(p)
end

function regGet(type, path)
    if type == "float" then
        return regGetFloat(path)
    elseif type == "int" then
        return regGetInt(path)
    elseif type == "bool" then
        return regGetBool(path)
    elseif type == "string" then
        return regGetString(path)
    end
    return nil
end
function regSet(type, path, value)
    if type == "float" then
        regSetFloat(path, value)
    elseif type == "int" then
        regSetInt(path, value)
    elseif type == "bool" then
        regSetBool(path, value)
    elseif type == "string" then
        regSetString(path, value)
    end
end