tool = {
    tool = {
        aimBody = nil,

        input = {
            didAttachDetachAI = function() return tool.tool.active() and InputPressed(regGetString('key.attach')) end,
            didCopyAI = function() return tool.tool.active() and InputPressed(regGetString('key.copy')) end,
            didActivateAI = function() return tool.tool.active() and InputPressed(regGetString('key.activate')) end,
            didConvert = function() return tool.tool.active() and InputPressed(regGetString('key.convert')) end,
            didStartStopTraining = function() return tool.tool.active() and InputPressed(regGetString('key.start')) end,
            didPauseResumeTraining = function() return tool.tool.active() and InputPressed(regGetString('key.pause')) end,
            didRemove = function() return tool.tool.active() and InputPressed(regGetString('key.remove')) end,
        },

        setup = {
            name = 'ai_trainer',
            title = 'AI Trainer',
            voxPath = 'MOD/vox/remote.vox',
        },

        active = function()
            return GetString('game.player.tool') == tool.tool.setup.name and (GetPlayerVehicle() == 0)
        end,

        init = function(enabled)
            RegisterTool(tool.tool.setup.name, tool.tool.setup.title, tool.tool.setup.voxPath)
            SetBool('game.tool.'..tool.tool.setup.name..'.enabled', enabled or true)
        end,

    }
}


function tool.run()

    local bcInfo = BodyController_info[1]

    -- Get cursor position at all times even if tool is not selected
    local playerTr = GetPlayerCameraTransform()
    local direction = TransformToParentVec(playerTr, Vec(0, 0, -1))
    local hit, dist, normal, shape = QueryRaycast(playerTr.pos, direction, 1000000)
    if hit then
        tool.tool.aimBody = GetMainBody(GetShapeBody(shape))
    else
        tool.tool.aimBody = nil
    end

    if  GetString("game.player.tool") == tool.tool.setup.name then
        if not UI_GAME then
            if tool.tool.input.didAttachDetachAI() then -- Attach/Detach AI to object
                if tool.tool.aimBody then
                    BodyController.instances.attachDetach(tool.tool.aimBody)
                end
            end
        
            if tool.tool.input.didCopyAI() then -- Copy AI from object
                if tool.tool.aimBody then
                    local bodyController = BodyController.instances.getBodyController(tool.tool.aimBody)
                    if bodyController then
                        regSetString('neural_network.current', ai.neural_network.ToJson(bodyController.neuralNetwork))
                    end
                end
            end

            if tool.tool.input.didActivateAI() then -- Activate/Deactivate target AI
                if tool.tool.aimBody then
                    local bodyController = BodyController.instances.getBodyController(tool.tool.aimBody)
                    if bodyController then
                        BodyController.toggleActivation(bodyController)
                    end
                end
            end

            if tool.tool.input.didConvert() then -- Convert all existing AIs to curent objective with current neural network
                local neuralNetwork = ai.neural_network.FromJson(regGetString('neural_network.current'))
                if not neuralNetwork then
                    dbm("Cannot convert all AIs: Syntax error in neural network")
                else
                    local objective = regGetString('neural_network.objective')
                    if not ai.neural_network.IsCompatible(neuralNetwork, objective) then
                        dbm("Cannot convert all AIs: Incompatible objective")
                    else
                        for i=1,#bcInfo.bodyControllersList do
                            BodyController.convert(bcInfo.bodyControllersList[i], objective, neuralNetwork)
                        end
                    end
                end
            end
        
            if tool.tool.input.didStartStopTraining() then -- Start/Stop training
                genetic_algo.toggleStart()
            end
        
            if tool.tool.input.didPauseResumeTraining() then -- Pause/Resume training
                genetic_algo.togglePause()
            end
        
            if tool.tool.input.didRemove() then -- Remove all AIs
                bcInfo.bodyControllersList = {}
            end
        end

        -- Move remote to correct location
        local t = Transform()
		t.pos = Vec(0.55, -0.55, -0.5)
		t.rot = QuatEuler(10, 0, 0)
		SetToolTransform(t)
    end

end