
table.unpack = table.unpack or unpack
--[[VECTORS]]
    -- Distance between two vectors.
    function VecDist(vec1, vec2) return VecLength(VecSub(vec1, vec2)) end
    -- Divide a vector by another vector.
    function VecDiv(v, n) n = n or 1 return Vec(v[1] / n, v[2] / n, v[3] / n) end
    -- Add all vectors in a table together.
    function VecAddAll(vectorsTable) local v = Vec(0,0,0) for i = 1, #vectorsTable do VecAdd(v, vectorsTable[i]) end return v end
    --- Returns a vector with random values.
    function rdmVec(min, max) return Vec(rdm(min, max),rdm(min, max),rdm(min, max)) end
    ---Print QuatEulers or vectors.
    function VecPrint(vec, decimals, label)
        DebugPrint(VecToStr(vec, decimals, label))
    end
    function VecToStr(vec, decimals, label)
        return (label or "") .. 
            "  " .. sfn(vec[1], decimals or 2) ..
            "  " .. sfn(vec[2], decimals or 2) ..
            "  " .. sfn(vec[3], decimals or 2)
    end
    function VecToTag(vec, decimals)
        return  sfn(vec[1], decimals or 2) ..
        "_" .. sfn(vec[2], decimals or 2) ..
        "_" .. sfn(vec[3], decimals or 2)
    end
    function VecFromTag(tag)
        local splitted = sSplit(tag, "_")
        return Vec(tonumber(splitted[1]), tonumber(splitted[2]), tonumber(splitted[3]))
    end
    function QuatToTag(quat, decimals)
        return  sfn(quat[1], decimals or 2) ..
        "_" .. sfn(quat[2], decimals or 2) ..
        "_" .. sfn(quat[3], decimals or 2) ..
        "_" .. sfn(quat[4], decimals or 2)
    end
    function QuatFromTag(tag)
        local splitted = sSplit(tag, "_")
        return Quat(tonumber(splitted[1]), tonumber(splitted[2]), tonumber(splitted[3]), tonumber(splitted[4]))
    end
    function TransformToTag(tr, decimals)
        return  VecToTag(tr.pos, decimals) ..
        "/" .. QuatToTag(tr.rot, decimals)
    end
    function TransformFromTag(tag)
        local splitted = sSplit(tag, "/")
        return Transform(VecFromTag(splitted[1]), QuatFromTag(splitted[2]))
    end
    -- Tells if two vectors face in the same direction
    function VecSameDir(vec1, vec2)
        return VecDot(vec1, vec2) >= 0
    end
    -- Align a vector to face in the same direction as a reference vector
    -- (negate its direction if needed)
    function VecAlign(vec, vecRef)
        return VecSameDir(vec, vecRef) and vec or VecScale(vec, -1)
    end
    -- Redirect a vector to face in the same direction as a reference vector
    function VecRedirect(vec, vecRef)
        return VecResize(vecRef, VecLength(vec))
    end
    -- Set the length of a vector without changing its orientation
    function VecResize(vec, size) return VecScale(VecNormalize(vec), size) end
    -- Project a vec1 on vec2 and return the length of vec1 projected
    function VecProjToVecLen(vec1, vec2) return VecDot(vec1, vec2) / VecLength(vec2) end
    -- Reflect a vector on a surface
    function VecReflect(vec, normal) return VecSub(vec, VecScale(normal, 2 * VecDot(vec, normal))) end
    -- Removed the vertical component of a vector
    function VecFlatten(vec) return Vec(vec[1], 0, vec[3]) end
    -- Gets the minimal angle in degree between two vectors
    function VecAngleBetween(vec1, vec2)return math.deg(math.acos(VecDot(vec1, vec2)/(VecLength(vec1)*VecLength(vec2)))) end
    -- Gets a 2D vector ignoring given axis
    function Vec3DTo2D(vec, axis)
        if axis == "x" then
            return {vec[2], vec[3]}
        elseif axis == "y" then
            return {vec[1], vec[3]}
        elseif axis == "z" then
            return {vec[1], vec[2]}
        end
        return nil
    end
    -- Gets the minimal angle in degree between two 2D vectors
    function Vec2DOrientedAngleBetween(vec1, vec2)
        local dot = vec1[1]*vec2[1] + vec1[2]*vec2[2]
        local det = vec1[1]*vec2[2] - vec1[2]*vec2[1]
        return math.deg(math.atan2(det, dot))
    end
    -- Rotate a vector on a given world axis
    function VecRotate(vec, axis, angle)
        local radAngle = math.rad(angle)
        if axis == "x" then
            local newY = (math.cos(radAngle)*vec[2])-(math.sin(radAngle)*vec[3])
            local newZ = (math.sin(radAngle)*vec[2])+(math.cos(radAngle)*vec[3])
            return Vec(vec[1] , newY, newZ)
        elseif axis == "y" then
            local newX = (math.cos(radAngle)*vec[1])-(math.sin(radAngle)*vec[3])
            local newZ = (math.sin(radAngle)*vec[1])+(math.cos(radAngle)*vec[3])
            return Vec(newX , vec[2], newZ)
        elseif axis == "z" then
            local newX = (math.cos(radAngle)*vec[1])-(math.sin(radAngle)*vec[2])
            local newY = (math.sin(radAngle)*vec[1])+(math.cos(radAngle)*vec[2])
            return Vec(newX , newY, vec[3])
        end
        return vec
    end



--[[QUAT]]
    function QuatLookDown(pos) return QuatLookAt(pos, VecAdd(pos, Vec(0, -1, 0))) end
    function QuatLookUp(pos) return QuatLookAt(pos, VecAdd(pos, Vec(0, 1, 0))) end
    function QuatTrLookDown(tr) return QuatLookAt(tr.pos, TransformToParentPoint(tr, Vec(0,-1,0))) end
    function QuatDir(dir) return QuatLookAt(Vec(0, 0, 0), dir) end -- Quat to 3d worldspace dir.
    function GetQuatEulerVec(quat) local x,y,z = GetQuatEuler(quat) return Vec(x,y,z) end
    function QuatConjugate(quat)
        return Quat(quat[1], -quat[2], -quat[3], -quat[4])
    end
    function QuatMagnitude(quat)
        return math.sqrt(math.pow(quat[1], 2) + math.pow(quat[2], 2) + math.pow(quat[3], 2) + math.pow(quat[4], 2))
    end
    function QuatInverse(quat)
        local conjugate = QuatConjugate(quat)
        local squaredMagnitude = math.pow(QuatMagnitude(quat), 2)
        return Quat(conjugate[1]/squaredMagnitude, conjugate[2]/squaredMagnitude, conjugate[3]/squaredMagnitude, conjugate[4]/squaredMagnitude)
    end
    function QuatWorldRotateQuat(a, b)
        -- Store current quat for later
        local initialQuat = QuatCopy(a)
        -- Negate current quat
        local tmpQuat = Quat()
        -- Apply new quat
        tmpQuat = QuatRotateQuat(tmpQuat, b)
        -- Re-apply old quat
        return QuatRotateQuat(tmpQuat, initialQuat)
    end
    function QuatWorldRotateVec(a, vec)
        -- Get quat from vec
        local initialQuat = QuatDir(vec)
        -- Negate quat on vec
        local tmpVec = QuatRotateVec(QuatInverse(initialQuat), vec)
        -- Get world rotated quat
        local tmpQuat = QuatWorldRotateQuat(initialQuat, a)
        -- Apply quat to vec
        return QuatRotateVec(tmpQuat, tmpVec)
    end



--[[AABB]]
    function AabbDraw(v1, v2, r, g, b, a)
        r = r or 1
        g = g or 1
        b = b or 1
        a = a or 1
        local x1 = v1[1]
        local y1 = v1[2]
        local z1 = v1[3]
        local x2 = v2[1]
        local y2 = v2[2]
        local z2 = v2[3]
        -- x lines top
        DebugLine(Vec(x1,y1,z1), Vec(x2,y1,z1), r, g, b, a)
        DebugLine(Vec(x1,y1,z2), Vec(x2,y1,z2), r, g, b, a)
        -- x lines bottom
        DebugLine(Vec(x1,y2,z1), Vec(x2,y2,z1), r, g, b, a)
        DebugLine(Vec(x1,y2,z2), Vec(x2,y2,z2), r, g, b, a)
        -- y lines
        DebugLine(Vec(x1,y1,z1), Vec(x1,y2,z1), r, g, b, a)
        DebugLine(Vec(x2,y1,z1), Vec(x2,y2,z1), r, g, b, a)
        DebugLine(Vec(x1,y1,z2), Vec(x1,y2,z2), r, g, b, a)
        DebugLine(Vec(x2,y1,z2), Vec(x2,y2,z2), r, g, b, a)
        -- z lines top
        DebugLine(Vec(x2,y1,z1), Vec(x2,y1,z2), r, g, b, a)
        DebugLine(Vec(x2,y2,z1), Vec(x2,y2,z2), r, g, b, a)
        -- z lines bottom
        DebugLine(Vec(x1,y1,z2), Vec(x1,y1,z1), r, g, b, a)
        DebugLine(Vec(x1,y2,z2), Vec(x1,y2,z1), r, g, b, a)
    end
    function AabbCheckOverlap(aMin, aMax, bMin, bMax)
        return 
        (aMin[1] <= bMax[1] and aMax[1] >= bMin[1]) and
        (aMin[2] <= bMax[2] and aMax[2] >= bMin[2]) and
        (aMin[3] <= bMax[3] and aMax[3] >= bMin[3])
    end
    function AabbCheckPointInside(aMin, aMax, p)
        return 
        (p[1] <= aMax[1] and p[1] >= aMin[1]) and
        (p[2] <= aMax[2] and p[2] >= aMin[2]) and
        (p[3] <= aMax[3] and p[3] >= aMin[3])
    end
    function AabbClosestEdge(pos, shape)

        local shapeAabbMin, shapeAabbMax = GetShapeBounds(shape)
        local bCenterY = VecLerp(shapeAabbMin, shapeAabbMax, 0.5)[2]
        local edges = {}
        edges[1] = Vec(shapeAabbMin[1], bCenterY, shapeAabbMin[3]) -- a
        edges[2] = Vec(shapeAabbMax[1], bCenterY, shapeAabbMin[3]) -- b
        edges[3] = Vec(shapeAabbMin[1], bCenterY, shapeAabbMax[3]) -- c
        edges[4] = Vec(shapeAabbMax[1], bCenterY, shapeAabbMax[3]) -- d

        local closestEdge = edges[1] -- find closest edge
        local index = 1
        for i = 1, #edges do
            local edge = edges[i]

            local edgeDist = VecDist(pos, edge)
            local closesEdgeDist = VecDist(pos, closestEdge)

            if edgeDist < closesEdgeDist then
                closestEdge = edge
                index = i
            end
        end
        return closestEdge, index
    end
    --- Sort edges by closest to startPos and closest to endPos. Return sorted table.
    function AabbSortEdges(startPos, endPos, edges)
        local s, startIndex = aabbClosestEdge(startPos, edges)
        local e, endIndex = aabbClosestEdge(endPos, edges)
        -- Swap first index with startPos and last index with endPos. Everything between stays same.
        edges = tableSwapIndex(edges, 1, startIndex)
        edges = tableSwapIndex(edges, #edges, endIndex)
        return edges
    end
    function AabbDimensions(min, max) return Vec(max[1] - min[1], max[2] - min[2], max[3] - min[3]) end
    function AabbGetShapeCenterPos(shape)
        local mi, ma = GetShapeBounds(shape)
        return VecLerp(mi,ma,0.5)
    end
    function AabbGetBodyCenterPos(body)
        local mi, ma = GetBodyBounds(body)
        return VecLerp(mi,ma,0.5)
    end
    function AabbGetShapeCenterTopPos(shape, addY)
        addY = addY or 0
        local mi, ma = GetShapeBounds(shape)
        local v =  VecLerp(mi,ma,0.5)
        v[2] = ma[2] + addY
        return v
    end
    function AabbGetBodyCenterTopPos(body, addY)
        addY = addY or 0
        local mi, ma = GetBodyBounds(body)
        local v =  VecLerp(mi,ma,0.5)
        v[2] = ma[2] + addY
        return v
    end


--[[OBB]]
function GetShapeObb(shape)
    local xsize, ysize, zsize, scale = GetShapeSize(shape)
    local min = Vec(0, 0, 0)
    local max = Vec(xsize * scale, ysize * scale, zsize * scale)
    local tr = GetShapeWorldTransform(shape)
    return tr, min, max
end
function GetBodyObb(body)
    local shapes = GetBodyShapes(body)
    local min = nil
    local max = nil
    local tr =  GetBodyTransform(body)
    -- print("##########################")
    for i=1,#shapes do
        local shape = shapes[i]
        local shapeTr, shapeMin, shapeMax = GetShapeObb(shape)
        -- print("shapePos="..VecStr(shapeTr.pos))
        -- print("shapeRot="..VecStr(GetQuatEulerVec(shapeTr.rot)))
        -- print("shapeMin="..VecStr(shapeMin))
        -- print("shapeMax="..VecStr(shapeMax))
        -- print("shapeSize="..VecStr(AabbDimensions(shapeMin, shapeMax)))
        -- print("-----")
        local shapeLocalTr = GetShapeLocalTransform(shape)
        local shapeLocalMin = TransformToParentPoint(shapeLocalTr, shapeMin)
        local shapeLocalMax = TransformToParentPoint(shapeLocalTr, shapeMax)
        -- print("shapeLocalPos="..VecStr(shapeLocalTr.pos))
        -- print("shapeLocalRot="..VecStr(GetQuatEulerVec(shapeLocalTr.rot)))
        -- print("shapeLocalMin="..VecStr(shapeLocalMin))
        -- print("shapeLocalMax="..VecStr(shapeLocalMax))
        -- print("shapeLocalSize="..VecStr(AabbDimensions(shapeLocalMin, shapeLocalMax)))
        -- print("=====")
        if min == nil then
            min = shapeLocalMin
            max = shapeLocalMax
        end
        min = Vec(
            math.min(min[1], shapeLocalMin[1], shapeLocalMax[1]),
            math.min(min[2], shapeLocalMin[2], shapeLocalMax[2]),
            math.min(min[3], shapeLocalMin[3], shapeLocalMax[3])
        )
        max = Vec(
            math.max(max[1], shapeLocalMin[1], shapeLocalMax[1]),
            math.max(max[2], shapeLocalMin[2], shapeLocalMax[2]),
            math.max(max[3], shapeLocalMin[3], shapeLocalMax[3])
        )
    end
    -- print("bodyPos="..VecStr(tr.pos))
    -- print("bodyRot="..VecStr(GetQuatEulerVec(tr.rot)))
    -- print("min="..VecStr(min))
    -- print("max="..VecStr(max))
    -- print("bodySize="..VecStr(AabbDimensions(min, max)))
    -- print("angleOffset="..VecStr(angleOffset))
    -- print("##########################")
    return tr, min, max
end
function GetObbCorners(tr, min, max)
    local corners = {}
    local localCorners = {
        Vec(min[1], min[2], min[3]),
        Vec(max[1], min[2], min[3]),
        Vec(max[1], min[2], max[3]),
        Vec(min[1], min[2], max[3]),
        Vec(min[1], max[2], min[3]),
        Vec(max[1], max[2], min[3]),
        Vec(max[1], max[2], max[3]),
        Vec(min[1], max[2], max[3]),
    }
    for i=1,#localCorners do
        corners[#corners+1] = TransformToParentPoint(tr, localCorners[i])
    end
    return corners
end
function ObbDraw(c, r, g, b, a)
    r = r or 1
    g = g or 1
    b = b or 1
    a = a or 1
    -- bottom square
    DebugLine(c[1], c[2], r, g, b, a)
    DebugLine(c[2], c[3], r, g, b, a)
    DebugLine(c[3], c[4], r, g, b, a)
    DebugLine(c[4], c[1], r, g, b, a)
    -- top square
    DebugLine(c[5], c[6], r, g, b, a)
    DebugLine(c[6], c[7], r, g, b, a)
    DebugLine(c[7], c[8], r, g, b, a)
    DebugLine(c[8], c[5], r, g, b, a)
    -- vertical segments
    DebugLine(c[1], c[5], r, g, b, a)
    DebugLine(c[2], c[6], r, g, b, a)
    DebugLine(c[3], c[7], r, g, b, a)
    DebugLine(c[4], c[8], r, g, b, a)
end


--[[TABLES]]
    function TableSwapIndex(t, i1, i2)
        local temp = t[i1]
        t[i1] = t[i2]
        t[i2] = temp
        return t
    end

    function TableClone(tb)
        local tbc = {}
        for k,v in pairs(tb) do tbc[k] = v end
        return tbc
    end

    function TableRemove(t, fnKeep)
        local j, n = 1, #t
    
        for i=1,n do
            if (fnKeep(t, i, j)) then
                -- Move i's kept value to j's position, if it's not already there.
                if (i ~= j) then
                    t[j] = t[i]
                    t[i] = nil
                end
                j = j + 1 -- Increment position of where we'll place the next kept value.
            else
                t[i] = nil
            end
        end
    
        return t
    end

    function TableAppend(t, elem)
        t[#t+1] = elem
    end

    function TableClear(t)
        for i=1,#t do t[i]=nil end
    end

    function DeepCopy(orig)
        local orig_type = type(orig)
        local copy
        if orig_type == 'table' then
            copy = {}
            for orig_key, orig_value in next, orig, nil do
                copy[DeepCopy(orig_key)] = DeepCopy(orig_value)
            end
            setmetatable(copy, DeepCopy(getmetatable(orig)))
        else
            copy = orig
        end
        return copy
    end

    function TableConcat(t1,t2)
        local t3 = {table.unpack(t1)}
        for i=1,#t2 do
            t3[#t3+1] = t2[i]
        end
        return t3
    end
    

    function TableEqual(t1, t2)
        -- Check length, or else the loop isn't valid.
        if #t1 ~= #t2 then
            return false
        end
        -- Check each element.
        for i, v in ipairs(t1) do
            if v ~= t2[i] then
                return false
            end
        end
        -- We've checked everything.
        return true
    end

    function TableFind(t, elem)
        -- Check each element.
        for i, v in ipairs(t) do
            if v == elem then
                return i
            end
        end
        -- We've checked everything.
        return nil
    end

    function TableBinarySearch(t, x)
        local lo = 1
        local hi = #t
        while lo < hi do
            local mid = math.floor((lo+hi) / 2)
            if t[mid] < x then 
                lo = mid+1
            else
                hi = mid
            end
        end
        return lo
     end

    function TableIndexOf(t, cond)
        local elem = nil
        local elemIndex = nil
        -- Check each element.
        for i, v in ipairs(t) do
            if elem == nil or cond(v, elem) then
                elem = v
                elemIndex = i
            end
        end
        -- We've checked everything.
        return elemIndex
    end

    function TableIndexOfMax(t)
        return TableIndexOf(t, function(elem, max) return elem > max end)
    end

    function TableIndexOfMin(t)
        return TableIndexOf(t, function(elem, min) return elem < min end)
    end

    function TablePushFront(t, elem, maxLength)
        maxLength = maxLength or #t
        -- Move each element one step back (remove last one)
        for i=maxLength-1,1,-1 do
            t[i+1]=t[i]
        end
        -- Add element first
        t[1] = elem
    end

    -- Multiply each value of the same index of two tables
    function TableMultiply(t1, t2)
        local res = {}
        for i, v in pairs(t2) do
            table.insert(res, t1[i] * t2[i])
        end
        return res
    end

    -- Transpose a matrix (2D table)
    function Matrixranspose(t)
        local transposed = {}
        for i=1,#t do
            local line = t[i]
            for j=1,#line do
                if not transposed[j] then
                    transposed[j] = {}
                end
                transposed[j][i] = line[j]
            end
        end
        return transposed
    end

    -- Multiple 2 matrices (2D tables)
    function MatrixMultiply(t1, t2)
        local multiplied = {}
        for i=1,#t1 do
            if not multiplied[i] then
                multiplied[i] = {}
            end
            for j=1,#t2[1] do
                if not multiplied[i][j] then
                    multiplied[i][j] = 0
                end
                for k=1,#t2 do
                    multiplied[i][j] = multiplied[i][j] + (t1[i][k] * t2[k][j])
                end
            end
        end
        return multiplied
    end
      
    



--[[RAYCASTING]]
---comment
---@param tr table
---@param distance number
---@param rad number
---@param rejectBodies table
---@param rejectShapes table
    function RaycastFromTransform(tr, distance, rad, rejectBodies, rejectShapes)

        if distance ~= nil then distance = -distance else distance = -300 end

        if rejectBodies ~= nil then for i = 1, #rejectBodies do QueryRejectBody(rejectBodies[i]) end end
        if rejectShapes ~= nil then for i = 1, #rejectShapes do QueryRejectShape(rejectShapes[i]) end end

        local plyTransform = tr
        local fwdPos = TransformToParentPoint(plyTransform, Vec(0, 0, distance))
        local direction = VecSub(fwdPos, plyTransform.pos)
        local dist = VecLength(direction)
        direction = VecNormalize(direction)
        local h, d, n, s = QueryRaycast(tr.pos, direction, dist, rad)
        if h then
            local p = TransformToParentPoint(plyTransform, Vec(0, 0, d * -1))
            local b = GetShapeBody(s)
            return h, p, s, b, d
        else
            return nil
        end
    end

    function QueryRequireTag(tag, value)
        if not queryRequiredTags then
            queryRequiredTags = {}
            queryRequiredValues = {}
        end
        queryRequiredTags[#queryRequiredTags] = tag
        queryRequiredValues[tag] = value
    end

    function QueryRejectTag(tag, value)
        if not queryRejectedTags then
            queryRejectedTags = {}
            queryRejectedValues = {}
        end
        queryRejectedTags[#queryRejectedTags] = tag
        queryRejectedValues[tag] = value
    end

    function QueryClearTags()
        queryRequiredTags = {}
        queryRequiredValues = {}
        queryRejectedTags = {}
        queryRejectedValues = {}
    end

    function HasRequiredTag(handle)
        if not queryRequiredTags then
            queryRequiredTags = {}
            queryRequiredValues = {}
        end
        for i=1,#queryRequiredTags do
            local tag = queryRequiredTags[i]
            local value = queryRequiredValues[tag]
            if HasTag(handle, tag) then
                if value ~= nil then
                    local valueFound = GetTagValue(handle, tag)
                    if valueFound ~= value then
                        return false
                    end
                end
            else
                -- For shapes, also check on parent body
                if GetEntityType(handle) == "shape" then
                    if not HasRequiredTag(GetShapeBody(handle)) then
                        return false
                    end
                else
                    return false
                end
            end
        end
        return true
    end

    function HasRejectedTag(handle)
        if not queryRejectedTags then
            queryRejectedTags = {}
            queryRejectedValues = {}
        end
        for i=1,#queryRejectedTags do
            local tag = queryRejectedTags[i]
            local value = queryRejectedValues[tag]
            if HasTag(handle, tag) then
                if value == nil then
                    return true
                else
                    local valueFound = GetTagValue(handle, tag)
                    if valueFound == value then
                        return true
                    end
                end
            else
                -- For shapes, also check on parent body
                if GetEntityType(handle) == "shape" then
                    if HasRejectedTag(GetShapeBody(handle)) then
                        return true
                    end
                end
            end
        end
        return false
    end

    function MatchesTagFilter(handle)
        return HasRequiredTag(handle) and not HasRejectedTag(handle)
    end

    function QueryXRaycast(origin, direction, maxDist, radius, rejectTransparent, rejectShapes)
        direction = VecNormalize(direction)
        if rejectShapes == nil then
            rejectShapes = {}
        end
        --local debugId = #rejectShapes
        for i=1,#rejectShapes do
            --dbp(debugId..": rejecting "..rejectShapes[i])
            QueryRejectShape(rejectShapes[i])
        end
        local hit, dist, normal, shape = QueryRaycast(origin, direction, maxDist, radius, rejectTransparent)
        --dbp(debugId..": QueryRaycast hit="..BoolToStr(hit))
        if hit then
            --dbp(debugId..": QueryRaycast dist="..dist..", normal="..VecStr(normal)..", shape="..shape)
            rejectShapes[#rejectShapes+1] = shape
            local dists, normals, shapes = QueryXRaycast(origin, direction, maxDist, radius, rejectTransparent, rejectShapes)
            if MatchesTagFilter(shape) then
                --dbp(debugId..": return matches filter")
                return {dist, table.unpack(dists)}, {normal, table.unpack(normals)}, {shape, table.unpack(shapes)}
            else
                --dbp(debugId..": return DOES NOT match filter")
                return dists, normals, shapes
            end
        end
        --dbp(debugId..": return empty")
        return {}, {}, {}
    end

    function QueryRaycastClosest(origin, direction, maxDist, radius, rejectTransparent, target, bestDistToTarget, rejectShapes)
        direction = VecNormalize(direction)
        if rejectShapes == nil then
            rejectShapes = {}
        end
        for i=1,#rejectShapes do
            QueryRejectShape(rejectShapes[i])
        end
        local hit, dist, normal, shape = QueryRaycast(origin, direction, maxDist, radius, rejectTransparent)
        if hit then
            rejectShapes[#rejectShapes+1] = shape
            local hitPoint = VecAdd(origin, VecResize(direction, dist))
            local distToTarget = VecDist(target, hitPoint)
            if bestDistToTarget == nil or distToTarget < bestDistToTarget then
                local hit2, dist2, normal2, shape2 = QueryRaycastClosest(origin, direction, maxDist, radius, rejectTransparent, target, distToTarget, rejectShapes)
                if hit2 then
                    if MatchesTagFilter(shape2) then
                        return hit2, dist2, normal2, shape2
                    end
                end
            else
                return false, dist, normal, shape
            end
        end
        if MatchesTagFilter(shape) then
            return hit, dist, normal, shape
        else
            return false, dist, normal, shape
        end
    end


--[[PHYSICS]]
    function diminishBodyAngVel(body, rate)
        local angVel = GetBodyAngularVelocity(body)
        local dRate = rate or 0.99
        local diminishedAngVel = Vec(angVel[1]*dRate, angVel[2]*dRate, angVel[3]*dRate)
        SetBodyAngularVelocity(body, diminishedAngVel)
    end
    function IsMaterialUnbreakable(mat, shape)
        return mat == 'rock' or mat == 'heavymetal' or mat == 'unbreakable' or mat == 'hardmasonry' or
            HasTag(shape,'unbreakable') or HasTag(GetShapeBody(shape),'unbreakable')
    end
    function IsBodyValid(body)
        return IsHandleValid(body) and (GetBodyMass(body) > 0)
    end
    function IsShapeValid(shape)
        return IsHandleValid(shape) and (GetShapeVoxelCount(shape) > 0)
    end
    function HighlightFullBody(body, r, g, b, a)
        local jointedBodies = GetJointedBodies(body)
        for i=1,#jointedBodies do
            DrawBodyOutline(jointedBodies[i], r, g, b, a)
        end
    end
    function GetMainBody(body)
        local jointedBodies = GetJointedBodies(body)
        local maxMass = GetBodyMass(body)
        local mainBody = body
        for i=1,#jointedBodies do
            local jointedBody = jointedBodies[i]
            local mass = GetBodyMass(jointedBody)
            if mass > maxMass then
                maxMass = mass
                mainBody = jointedBody
            end
        end
        return mainBody
    end
    function GetMainShape(body)
        local shapes = GetBodyShapes(body)
        local maxVoxels = 0
        local mainShape = nil
        for i=1,#shapes do
            local shape = shapes[i]
            local voxels = GetShapeVoxelCount(shape)
            if voxels > maxVoxels then
                maxVoxels = voxels
                mainShape = shape
            end
        end
        return mainShape
    end



--[[VFX]]
    colors = {
        white = Vec(1,1,1),
        black = Vec(0,0,0),
        grey = Vec(0,0,0),
        red = Vec(1,0,0),
        blue = Vec(0,0,1),
        yellow = Vec(1,1,0),
        purple = Vec(1,0,1),
        green = Vec(0,1,0),
        orange = Vec(1,0.5,0),
    }
    function DrawDot(pos, l, w, r, g, b, a, dt)
        local dot = LoadSprite("ui/hud/dot-small.png")
        local spriteRot = QuatLookAt(pos, GetCameraTransform().pos)
        local spriteTr = Transform(pos, spriteRot)
        if dt == nil then dt = true end
        DrawSprite(dot, spriteTr, l or 0.2, w or 0.2, r or 1, g or 1, b or 1, a or 1, dt and true)
    end



--[[SOUND]]
    local debugSounds = {
        beep = LoadSound("warning-beep"),
        buzz = LoadSound("light/spark0"),
        chime = LoadSound("elevator-chime"),
        valu = LoadSound("valuable.ogg"),}
    function beep(pos, vol) PlaySound(debugSounds.beep, pos or GetCameraTransform().pos, vol or 0.3) end
    function buzz(pos, vol) PlaySound(debugSounds.buzz, pos or GetCameraTransform().pos, vol or 0.3) end
    function chime(pos, vol) PlaySound(debugSounds.chime, pos or GetCameraTransform().pos, vol or 0.3) end
    function valu(pos, vol) PlaySound(debugSounds.valu, pos or GetCameraTransform().pos, vol or 0.3) end



--[[MATH]]
    function round(n, dec) local pow = 10^dec return math.floor(n * pow) / pow end
    --- return number if > 0, else return 0.00000001
    function gtZero(n) if n <= 0 then return 0.00000001 end return n end
    --- return number if not = 0, else return 0.00000001
    function nZero(n) if n == 0 then return 0.00000001 end return n end
    --- return random float betzeen min and max
    function rdmf(min, max)
        min = min or 0
        max = max or 1
        return min + (math.random() * (max-min))
    end
    --- return random int betzeen min and max
    function rdm(min, max) return math.random(min or 0, max or 1) end
    --- return a random number following a standard normal distribution (mean of 0 and standard deviation of 1)
    function rdmsd()
        return math.sqrt(-2*math.log(rdmf()))*math.cos(2*math.pi*rdmf())
    end
    function clamp(value, mi, ma)
        if value < mi then value = mi end
        if value > ma then value = ma end
        return value
    end
    -- Brings a value from range [mi, ma] to range [0, 1]
    function ratio(value, mi, ma)
        return (value - mi) / (ma - mi)
    end
    function oscillate(time)
        local a = (GetTime() / (time or 1)) % 1
        a = a * math.pi
        return math.sin(a)
    end
    -- return a random point on the surface of a sphere centered on center and of radius radius
    function getRandPointOnSphere(center, radius)
        local lambda = rdmf(-180, 180)
        local phi = math.acos(2 * rdmf() - 1)
        local x = math.cos(lambda) * math.cos(phi)
        local y = math.cos(lambda) * math.sin(phi)
        local z = math.sin(lambda)
        return VecAdd(center, Vec(x * radius, y * radius, z * radius))
    end
    function floatToInt(float)
        return math.floor(float + .5)
    end
    function getMinDistBetweenSpherePoints(transform, radius, samples)
		local points = {}
		local sqrt, sin, cos = math.sqrt, math.sin, math.cos
		radius = radius or 1
		samples = samples or 100

		-- Converted from python to lua, see original code https://stackoverflow.com/a/26127012/5459461
		local points = {}
		for i = samples-2, samples-1 do
			local y = 1 - (i / (samples - 1)) * 2
			local rad = sqrt( 1 - y * y )
			local theta = 2.399963229728653 * i

			local x = cos( theta ) * rad
			local z = sin( theta ) * rad
			local point = TransformToParentPoint( Transform( transform.pos, transform.rot),
			                                      Vec( x * radius, y * radius, z * radius ) )
			TableAppend(points, point)
		end
        local dist = VecDist(points[1], points[2])
        -- dbp("getMinDistBetweenSpherePoints radius="..radius..", samples="..samples..", dist="..dist)
		return dist
	end


--[[LOGIC]]
    function ternary ( cond , T , F )
        if cond then return T else return F end
    end



--[[FORMATTING]]
    --- string format. default 2 decimals.
    function sfn(numberToFormat, dec)
        local s = (tostring(dec or 2))
        return string.format("%."..s.."f", numberToFormat)
    end
    function sfnTime(dec) return sfn(' '..GetTime(), dec or 4) end
    function sfnCommas(dec)
        return tostring(math.floor(dec)):reverse():gsub("(%d%d%d)","%1,"):gsub(",(%-?)$","%1"):reverse()
        -- https://stackoverflow.com/questions/10989788/format-integer-in-lua
    end

    function sSplit(inputstr, sep)
        if sep == nil then
            sep = "%s"
        end
        local t={}
        for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
            table.insert(t, str)
        end
        return t
    end
    function BoolToStr(bool)
        return bool and "true" or "false"
    end




--[[TIMERS]]

    ---Run a timer and a table of functions.
    ---@param timer table -- = {time, delay}
    ---@param functions table -- Table of functions that are called when time = 0.
    ---@param runTime boolean -- Decrement time when calling this function.
    function TimerRunTimer(timer, functions, runTime)
        if timer.time <= 0 then
            TimerResetTime(timer)

            for i = 1, #functions do
                functions[i]()
            end

        elseif runTime then
            TimerRunTime(timer)
        end
    end

    -- Only runs the timer countdown if there is time left.
    function TimerRunTime(timer)
        if timer.time > 0 then
            timer.time = timer.time - GetTimeStep()
        end
    end

    -- Set time left to 0.
    function TimerEndTime(timer)
        timer.time = 0
    end

    -- Reset time to start delay.
    function TimerResetTime(timer)
        timer.time = timer.delay
    end

--[[PLAYER]]

    -- Get center point of the player
    function GetPlayerCenter()
        local playerEyeTr = GetPlayerCameraTransform()
        local playerFeetTr = GetPlayerTransform()
        return VecAdd(playerFeetTr.pos, VecScale(VecSub(playerEyeTr.pos, playerFeetTr.pos), 0.5))
    end